# -*- coding: utf-8 -*-
from collections import MutableMapping

from .pyjmad import cern, java


class Element(object):
    def __init__(self, jmadElement):
        self._jmadElement = jmadElement

    @property
    def name(self):
        return self._jmadElement.getName()

    @property
    def type(self):
        return str(self._jmadElement.getMadxElementType())

    @property
    def length(self):
        return self._jmadElement.getLength()

    @length.setter
    def length(self, len):
        self._jmadElement.getLength(float(len))

    @property
    def position(self):
        return self._jmadElement.getPosition().getValue()

    @position.setter
    def position(self, position):
        self._jmadElement.setPosition(float(position))

    @property
    def attributes(self):
        return Attributes(self._jmadElement)

    def __str__(self):
        return self.name + ' (' + self.type + ')'

    def __repr__(self):
        return self.name + ' (' + self.type + ': ' + str(self.attributes) + ')'


def _specific_element(name, attributes):
    attr_dict = {}
    for attr in attributes:
        java_name = ''.join([s.capitalize() for s in attr.split('_')])
        p = property(lambda self, java_name=java_name: self._jmadElement.__getattribute__('get' + java_name)()) \
            .setter(
            lambda self, v, java_name=java_name: self._jmadElement.__getattribute__('set' + java_name)(float(v)))
        attr_dict[attr] = p
    return type(name, (Element,), attr_dict)


BeamBeam = _specific_element('BeamBeam',
                             ['charge', 'direction', 'displacement_x', 'displacement_y', 'shape', 'sig_x', 'sig_y',
                              'width'])
Bend = _specific_element('Bend',
                         ['angle', 'e1', 'e2', 'fint', 'fint_x', 'h1', 'h2', 'h_gap', 'k0', 'k1', 'k2', 'k3', 'tilt'])
Corrector = _specific_element('Corrector', ['h_kick', 'tilt', 'v_kick'])
Marker = _specific_element('Marker', [])
Monitor = _specific_element('Monitor', [])
Quadrupole = _specific_element('Quadrupole', ['k1', 'tilt'])
Sextupole = _specific_element('Sextupole', ['k2', 'tilt'])
Octupole = _specific_element('Octupole', ['k3', 'tilt'])


def from_jmad(jmadElement):
    _jmadElementMap = {
        'BeamBeam': BeamBeam,
        'Bend': Bend,
        'Corrector': Corrector,
        'Marker': Marker,
        'Monitor': Monitor,
        'Octupole': Octupole,
        'Quadrupole': Quadrupole,
        'Sextupole': Sextupole,
        'UnknownElement': Element
    }
    return _jmadElementMap[jmadElement.getClass().getSimpleName()](jmadElement)


class Attributes(MutableMapping):
    def __init__(self, jmadElement):
        self._jmadElement = jmadElement

    def __getitem__(self, k):
        return self._jmadElement.getAttribute(k).doubleValue()

    def __setitem__(self, k, v):
        return self._jmadElement.setAttribute(k, java.lang.Double(float(v)))

    def __delitem__(self, k):
        raise NotImplementedError("Deletion of Attributes is not supported")

    def __iter__(self):
        for s in self._jmadElement.getAttributeNames():
            yield s

    def __len__(self):
        return self._jmadElement.getAttributeNames().size()

    def _ipython_key_completions_(self):
        return [s for s in self._jmadElement.getAttributeNames()]

    def __repr__(self):
        return '{' + ', '.join([k + '=' + str(v) for k, v in self.items()]) + '}'

    def __str__(self):
        return self.__repr__()



__version__ = "0.0.5"

__cmmnbuild_deps__ = [
    "jmad-core-pro",
    "jmad-gui-pro",
    "slf4j-log4j12",
    "slf4j-api",
    "log4j"
]

__gradle_deps__ = [
    "jmad:jmad-core:0.1.4+",
    'jmad:jmad-modelpack-service:0.2.3+',
    'jmad:jmad-modelpack-fx:0.1.1+',
    "jmad:jmad-gui:0.1.1+",
    "org.slf4j:slf4j-api:+",
    "org.slf4j:slf4j-log4j12:+",
    "log4j:log4j:1.2.17",
    # for jdataviewer
    {'repository': 'ivy { url "http://jmad.web.cern.ch/jmad/static/ws/lib/"\n' +
                   'layout "pattern", { artifact "[artifact].[ext]" } \n' +
                   'metadataSources { artifact() } }',
     'groupId': 'cern', 'artifactId': 'jdataviewer', 'version': '1.4.7'},
]

# When running setuptools without required dependencies installed
# we need to be able to access __version__, so print a warning but
# continue
try:
    from .pyjmad import *
    from . import element, matching
except:
    import logging

    logging.basicConfig()
    log = logging.getLogger(__name__)
    log.exception("required dependencies are not installed")



# -*- coding: utf-8 -*-

class GlobalConstraint(object):
    def __init__(self, Q1=None, Q2=None, DQ1=None, DQ2=None):
        self.Q1 = Q1
        self.Q2 = Q2
        self.DQ1 = DQ1
        self.DQ2 = DQ2

    @property
    def constraints(self):
        return {k: v for k, v in self.__dict__.items() if v is not None}

    def __repr__(self):
        return repr(self.constraints)


class LocalConstraint(object):
    def __init__(self, at,
                 ALFX=None, ALFY=None,
                 BETX=None, BETY=None,
                 DDPX=None, DDPY=None,
                 DDX=None, DDY=None,
                 DPX=None, DPY=None,
                 DX=None, DY=None,
                 MUX=None, MUY=None,
                 PX=None, PY=None,
                 X=None, Y=None):
        self.element = at
        self.ALFX = ALFX
        self.ALFY = ALFY
        self.BETX = BETX
        self.BETY = BETY
        self.DDPX = DDPX
        self.DDPY = DDPY
        self.DDX = DDX
        self.DDY = DDY
        self.DPX = DPX
        self.DPY = DPY
        self.DX = DX
        self.DY = DY
        self.MUX = MUX
        self.MUY = MUY
        self.PX = PX
        self.PY = PY
        self.X = X
        self.Y = Y

    @property
    def constraints(self):
        return {k: v for k, v in self.__dict__.items() if k != 'element' and v is not None}

    def __repr__(self):
        return self.element + ' -> ' + repr(self.constraints)


class Vary(object):
    def __init__(self, strength, lower_bound=None, upper_bound=None, step=None):
        self.strength = strength
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.step = step

    def __repr__(self):
        return 'Vary(' + self.strength + ' -> ' + str(self.lower_bound) + ':' + str(self.upper_bound) + ':' + str(
            self.step) + ')'



import itertools
import logging

from .pyjmad import org, java
from .util import *


def _repo_to_uri(javaRepo):
    if str(javaRepo.connectorId()) == 'internal-classpath':
        return 'jmad:internal'
    elif str(javaRepo.connectorId()) == 'gitlab-group-api-v4':
        return javaRepo.baseUrl() + '/' + javaRepo.repoName()
    else:
        raise ValueError('unknown repository type: ' + str(javaRepo))


def _repo_from_uri(uri):
    JMadModelPackageRepository = org.jmad.modelpack.domain.JMadModelPackageRepository
    InternalRepository = org.jmad.modelpack.connect.embedded.domain.InternalRepository
    schema = uri.split(':')[0]
    if schema == 'jmad':
        if uri == 'jmad:internal':
            return InternalRepository.INTERNAL
        else:
            raise ValueError('invalid "jmad:" uri "{0}" - expected = jmad:internal'.format(uri))
    elif schema == 'http' or schema == 'https':
        base_url, group_name = uri.rsplit('/', 1)
        return JMadModelPackageRepository(base_url, group_name, 'gitlab-group-api-v4')
    elif '+' in schema:
        connector, repo_uri = uri.split('+', 1)
        base_url, group_name = repo_uri.rsplit('/', 1)
        return JMadModelPackageRepository(base_url, group_name, connector)
    else:
        raise ValueError('invalid repository URI: ' + uri)


jmad_default_repositories = {}
try:
    _JMadModelRepositories = org.jmad.modelpack.domain.JMadModelRepositories
    for method in _JMadModelRepositories.__javaclass__.getDeclaredMethods():
        if method.getModifiers() & java.lang.reflect.Modifier.PUBLIC:
            repo = method.getName()
            jmad_default_repositories[repo] = _repo_to_uri(getattr(_JMadModelRepositories, repo)())
except Exception:
    logging.exception("can not fetch default models from jmad-modelpack-service")

class JMadModelPackService(object):
    def __init__(self, applicationContext):
        self._javaService = applicationContext['jmadModelPackageService']
        self._javaRepositoryManager = applicationContext['packageRepositoryManager']
        self._modelpacks = {}
        self._reload()

    def _reload(self):
        def modelpack_name(pkg):
            return str(pkg.modelPackage().name())

        logging.info('Fetching available ModelPacks from ' + str(self.repositories))
        self._modelpacks = {}
        modelpacks = list(self._javaService.availablePackages().collectList().block())
        modelpacks.sort(key=modelpack_name)
        for name, variants in itertools.groupby(modelpacks, modelpack_name):
            self._modelpacks[name] = ModelPackType(self, name, variants)

    @property
    def repositories(self):
        return [_repo_to_uri(r) for r in self._javaRepositoryManager.enabledRepositories().collectList().block()]

    def add_repository(self, repo_uri):
        self._javaRepositoryManager.enable(_repo_from_uri(repo_uri))
        self._reload()

    def remove_repository(self, repo_uri):
        self._javaRepositoryManager.disable(_repo_from_uri(repo_uri))
        self._reload()

    def __getitem__(self, modelpack):
        return self._modelpacks[modelpack]

    def __setitem__(self, k, v):
        raise NotImplementedError('Setting of ModelPacks is not supported')

    def __delitem__(self, k):
        raise NotImplementedError('Deletion of ModelPacks is not supported')

    def __iter__(self):
        return self._modelpacks.__iter__()

    def __len__(self):
        return self._modelpacks.__len__()

    def __repr__(self):
        return 'ModelPackService [ \n' \
               ' -> Enabled Repos: \n    - ' + '\n    - '.join(self.repositories) + '\n' + \
               ' -> Available ModelPacks: \n    - ' + '\n    - '.join(self._modelpacks.keys()) + '\n' + \
               ']'

    def __str__(self):
        return self.__repr__()

    def _ipython_key_completions_(self):
        return list(self._modelpacks.keys())


class ModelPackType(object):
    def __init__(self, service, name, variants):
        self._name = name
        self._all_variants = {}
        for modpack in variants:
            if str(modpack.variant().type()) == 'RELEASE':
                variant = 'releases'
            elif str(modpack.variant().type()) == 'BRANCH':
                variant = 'branches'
            elif str(modpack.variant().type()) == 'TAG':
                variant = 'tags'
            elif type(modpack.variant()) is org.jmad.modelpack.connect.embedded.domain.InternalPackageVariant:
                variant = 'internal_models'
            else:
                variant = 'others'
            if variant == 'internal_models':
                self._all_variants[variant] = modpack
            else:
                self._all_variants.setdefault(variant, {})[modpack.variant().name()] = ModelPack(service, modpack)

    def __dir__(self):
        return ['name'] + list(self._all_variants.keys())

    def __getattr__(self, name):
        if name == 'name':
            return self._name
        elif name in self._all_variants:
            return self._all_variants[name]
        else:
            raise AttributeError


class ModelPack(object):
    def __init__(self, service, javaModelPackVariant):
        self._javaModelPackVariant = javaModelPackVariant
        self._service = service

    @property
    def id(self):
        return str(self._javaModelPackVariant.fullName())

    @property
    def models(self):
        from .pyjmad import ModelDefinition
        modeldefs = self._service._javaService.modelDefinitionsFrom(self._javaModelPackVariant).collectList().block()
        return HtmlDict({mdef.getName(): ModelDefinition(mdef) for mdef in modeldefs})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import setuptools
from setuptools.command.install import install as _install

import pyjmad


class install(_install):
    def run(self):
        print('registering with dependency manager')
        import cmmnbuild_dep_manager
        mgr = cmmnbuild_dep_manager.Manager()
        mgr.install('pyjmad')
        print('registered pyjmad with gradle_dep_manager')
        _install.run(self)


setuptools.setup(
    name='pyjmad',
    version=pyjmad.__version__,
    description='A Python wrapping of JMad',
    author='Michi Hostettler',
    author_email='michi.hostettler@cern.ch',
    url='https://github.com/pymad/pyjmad',
    packages=['pyjmad'],
    install_requires=['JPype1>=0.6.1',
                      'cmmnbuild-dep-manager>=2.2.0',
                      'numpy', 'pandas'],
    cmdclass={'install': install},
)



import jpype

org = jpype.JPackage('org')
AnnotationConfigApplicationContext = org.springframework.context.annotation.AnnotationConfigApplicationContext
ClassPathXmlApplicationContext = org.springframework.context.support.ClassPathXmlApplicationContext

class SpringApplicationContext(object):
    def __init__(self, configuration):
        if type(configuration) is str:
            self._context = ClassPathXmlApplicationContext(configuration)
        else:
            self._context = AnnotationConfigApplicationContext(configuration)

    def bean_definitions(self):
        return [str(s) for s in self._context.getBeanDefinitionNames()]

    def __getitem__(self, bean):
        return self._context.getBean(bean)

    def __setitem__(self, k, v):
        raise NotImplementedError('Setting of Beans is not supported')

    def __delitem__(self, k):
        raise NotImplementedError('Deletion of Beans is not supported')

    def __iter__(self):
        for s in self.bean_definitions():
            yield self.__getitem__(s)

    def __len__(self):
        return len(self.bean_definitions())

    def __repr__(self):
        return str(self._context.toString())

    def __str__(self):
        return str(self._context.toString())

    def _ipython_key_completions_(self):
        return self.bean_definitions()



# -*- coding: utf-8 -*-
from collections import namedtuple, MutableMapping, Mapping, OrderedDict

import numpy as np
import pandas as pd
import re, site, logging

import cmmnbuild_dep_manager

mgr = cmmnbuild_dep_manager.Manager('pyjmad')

jpype = mgr.start_jpype_jvm()

cern = jpype.JPackage('cern')
org = jpype.JPackage('org')
java = jpype.JPackage('java')
com = jpype.JPackage('com')

org.apache.log4j.BasicConfigurator.configure()
org.apache.log4j.Logger.getRootLogger().setLevel(org.apache.log4j.Level.INFO)
from .spring import SpringApplicationContext

from .modelpack import JMadModelPackService
from .util import *

JMadServiceFactory = cern.accsoft.steering.jmad.service.JMadServiceFactory
Doubles = com.google.common.primitives.Doubles
JMadGui = cern.accsoft.steering.jmad.gui.JMad
JMadGuiPreferencesImpl = cern.accsoft.steering.jmad.gui.manage.impl.JMadGuiPreferencesImpl
TfsResultRequestImpl = cern.accsoft.steering.jmad.domain.result.tfs.TfsResultRequestImpl
Iterables = com.google.common.collect.Iterables


class JMad(object):
    def __init__(self):
        try:
            try:
                self._springContext = SpringApplicationContext(
                    cern.accsoft.steering.jmad.gui.config.JMadGuiStandaloneConfiguration)
            except:
                logging.info("Could not instantiate GUI context, trying headless context")
                self._springContext = SpringApplicationContext(
                    org.jmad.modelpack.service.conf.JMadModelPackageServiceStandaloneConfiguration)

            self._jmadService = self._springContext['jmadService']
            self.model_packs = JMadModelPackService(self._springContext)
        except:
            logging.exception("JMad Model Pack Service not available, falling back")
            self._jmadService = JMadServiceFactory.createJMadService()
        self._jmadModelDefinitionManager = self._jmadService.getModelDefinitionManager()
        self._jmadModelManager = self._jmadService.getModelManager()

    @property
    def model_definitions(self):
        return HtmlDict({m.getName(): ModelDefinition(m) for m in
                         self._jmadModelDefinitionManager.getAllModelDefinitions()})

    def create_model(self, model):
        if type(model) is str:
            model = self.model_definitions[model]
        elif type(model) is not ModelDefinition:
            model = ModelDefinition(model)
        return Model(self._jmadService.createModel(model._jmadModelDefinition))

    def open_jmad_gui(self):
        gui = self._springContext['jmadGui']
        gui.getJmadGuiPreferences().setCleanupOnClose(False)
        gui.getJmadGuiPreferences().setExitOnClose(False)
        gui.getJmadGuiPreferences().setMainFrame(False)
        jpype.setupGuiEnvironment(lambda: gui.show())


class Model(object):
    def __init__(self, jmadModel):
        self._jmadModel = jmadModel
        self.definition = ModelDefinition(jmadModel.getModelDefinition())
        if not jmadModel.isInitialized():
            jmadModel.init()

    def twiss(self, variables, element_filter=None):
        if isinstance(variables, str):
            variables = [variables, ]
        request = TfsResultRequestImpl.createDefaultRequest()
        for var in variables:
            request.addVariable(MadxTwissVariable._jmad_variable(var))
        if element_filter is not None:
            request.addElementFilter(element_filter)
        tfs_result = self._jmadModel.twiss(request)
        return _jmad_TfsResult_to_pandas(tfs_result)

    @property
    def name(self):
        return self.definition.name

    @property
    def optic(self):
        jmadOptic = self._jmadModel.getActiveOpticsDefinition()
        if jmadOptic is None:
            return None
        else:
            return jmadOptic.getName()

    @optic.setter
    def optic(self, optic):
        jmadOptic = self.definition._jmad_opticDefinition(optic)
        if jmadOptic is None:
            raise ValueError('Invalid Optic: ' + jmadOptic)
        self._jmadModel.setActiveOpticsDefinition(jmadOptic)

    @property
    def sequence(self):
        jmadRange = self._jmadModel.getActiveRange()
        if jmadRange is None:
            return None
        else:
            return SequenceDefinition(jmadRange.getRangeDefinition().getSequenceDefinition())

    @sequence.setter
    def sequence(self, sequence):
        jmadSequence = self.definition._jmad_sequenceDefinition(sequence)
        if jmadSequence is None:
            raise ValueError('Invalid sequence: ' + sequence)
        self._jmadModel.setActiveRangeDefinition(jmadSequence.getDefaultRangeDefinition())

    @property
    def range(self):
        return self._jmadModel.getActiveRange().getName()

    @range.setter
    def range(self, range):
        return self._jmadModel.setActiveRangeDefinition(self.sequence._jmad_rangeDefinition(range))

    @property
    def strengths(self):
        return Strengths(self._jmadModel.getStrengthsAndVars())

    @property
    def elements(self):
        return Elements(self._jmadModel.getActiveRange())

    @property
    def beam(self):
        return Beam(self._jmadModel.getActiveRange().getRangeDefinition().getSequenceDefinition().getBeam())

    def reset(self):
        self._jmadModel.reset()

    def match(self, *args):
        MatchResultRequestImpl = cern.accsoft.steering.jmad.domain.result.match.MatchResultRequestImpl
        MadxVaryParameterImpl = cern.accsoft.steering.jmad.domain.result.match.input.MadxVaryParameterImpl
        MatchConstraintGlobal = cern.accsoft.steering.jmad.domain.result.match.input.MatchConstraintGlobal
        MatchConstraintLocal = cern.accsoft.steering.jmad.domain.result.match.input.MatchConstraintLocal
        MadxParameterImpl = cern.accsoft.steering.jmad.domain.knob.MadxParameterImpl
        MadxRange = cern.accsoft.steering.jmad.domain.machine.MadxRange
        from .matching import LocalConstraint, GlobalConstraint, Vary
        jmadRequest = MatchResultRequestImpl()
        for arg in args:
            if isinstance(arg, LocalConstraint):
                jmadConstraint = MatchConstraintLocal(MadxRange(arg.element))
                for c, v in arg.constraints.items():
                    jmadConstraint.__getattribute__('set' + (c.capitalize()))(java.lang.Double(float(v)))
                jmadRequest.addMatchConstraint(jmadConstraint)
            elif isinstance(arg, GlobalConstraint):
                jmadConstraint = MatchConstraintGlobal()
                for c, v in arg.constraints.items():
                    jmadConstraint.__getattribute__('set' + (c.capitalize()))(java.lang.Double(float(v)))
                jmadRequest.addMatchConstraint(jmadConstraint)
            elif isinstance(arg, Vary):
                jmadVary = MadxVaryParameterImpl(MadxParameterImpl(arg.strength))
                if arg.lower_bound is not None:
                    jmadVary.setLower(java.lang.Double(float(arg.lower_bound)))
                if arg.upper_bound is not None:
                    jmadVary.setUpper(java.lang.Double(float(arg.upper_bound)))
                if arg.step is not None:
                    jmadVary.setStep(java.lang.Double(float(arg.step)))
                jmadRequest.addMadxVaryParameter(jmadVary)
            else:
                raise ValueError('Expecting LocalConstraint, GlobalConstraint, Vary - but got ' + str(type(arg)))

        return MatchResult(self._jmadModel.match(jmadRequest))

    def __str__(self):
        return self.name + ' - ' + str(self.optic) + ' - ' + str(self.sequence) + ' - ' + str(self.range)

    def __repr__(self):
        return self.__str__()


class ModelDefinition(object):
    def __init__(self, jmadModelDefinition):
        self._jmadModelDefinition = jmadModelDefinition

    @property
    def name(self):
        return self._jmadModelDefinition.getName()

    @property
    def optics(self):
        return [o.getName() for o in self._jmadModelDefinition.getOpticsDefinitions()]

    @property
    def sequences(self):
        return [SequenceDefinition(o) for o in self._jmadModelDefinition.getSequenceDefinitions()]

    def _jmad_opticDefinition(self, optic):
        if isinstance(optic, str):
            return self._jmadModelDefinition.getOpticsDefinition(optic)
        else:
            return optic

    def _jmad_sequenceDefinition(self, sequence):
        if isinstance(sequence, str):
            return self._jmadModelDefinition.getSequenceDefinition(sequence)
        elif isinstance(sequence, SequenceDefinition):
            return sequence._jmadSequenceDefinition
        else:
            return sequence

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name + ' (optics=' + str(self.optics) + ', sequences=' + str(self.sequences) + ')'

    def _repr_html_(self):
        html = '<table><thead>'
        html += '<tr><th colspan="2"><em>' + self.name + '</em></th></tr>'
        html += '<tr><th>Optics</th><th>Sequences</th></tr>'
        html += '</thead><tbody><tr>'
        html += '<td style="vertical-align:top"><ul>'
        for r in self.optics:
            html += '<li>' + r + '</li>'
        html += '</ul></td>'
        html += '<td style="vertical-align:top"><ul>'
        for s in self.sequences:
            html += '<li>' + s._repr_html_() + '</li>'
        html += '</ul></td>'
        html += '</tr></tbody></table>'
        return html


class Beam(object):
    def __init__(self, jmadBeam):
        self._jmadBeam = jmadBeam

    @property
    def bunch_current(self):
        return _unbox_double(self._jmadBeam.getBunchCurrent())

    @property
    def bunch_length(self):
        return _unbox_double(self._jmadBeam.getBunchLength())

    @property
    def bunch_number(self):
        return _unbox_integer(self._jmadBeam.getBunchNumber())

    @property
    def is_bunched(self):
        return _unbox_boolean(self._jmadBeam.getBunched())

    @property
    def charge(self):
        return _unbox_double(self._jmadBeam.getCharge())

    @property
    def direction(self):
        v = self._jmadBeam.getDirection()
        if str(v) == 'PLUS':
            return +1
        elif str(v) == 'MINUS':
            return -1
        else:
            return None

    @property
    def energy(self):
        return _unbox_double(self._jmadBeam.getEnergy())

    @property
    def gamma(self):
        return _unbox_double(self._jmadBeam.getGamma())

    @property
    def emittance_pysical_h(self):
        return _unbox_double(self._jmadBeam.getHorizontalEmittance())

    @property
    def emittance_pysical_v(self):
        return _unbox_double(self._jmadBeam.getVerticalEmittance())

    @property
    def mass(self):
        return _unbox_double(self._jmadBeam.getMass())

    @property
    def momentum(self):
        return _unbox_double(self._jmadBeam.getMomentum())

    @property
    def emittance_normalized_h(self):
        return _unbox_double(self._jmadBeam.getNormalisedHorizontalEmittance())

    @property
    def emittance_normalized_v(self):
        return _unbox_double(self._jmadBeam.getNormalisedVerticalEmittance())

    @property
    def particle(self):
        v = self._jmadBeam.getParticle()
        if v is None:
            return None
        else:
            return str(v)

    @property
    def radiate(self):
        return _unbox_boolean(self._jmadBeam.getRadiate())

    @property
    def relative_energy_spread(self):
        return _unbox_double(self._jmadBeam.getRelativeEnergySpread())

    def __str__(self):
        return self.particle + ' beam @ ' + str(self.energy) + ' GeV'

    def __repr__(self):
        s = self.__str__() + '\n'
        for f in dir(self):
            if not f.startswith('_') and self.__getattribute__(f) is not None:
                s += '   ' + f + ' = ' + str(self.__getattribute__(f)) + '\n'
        return s


class SequenceDefinition(object):
    def __init__(self, jmadSequenceDefinition):
        self._jmadSequenceDefinition = jmadSequenceDefinition

    def _jmad_rangeDefinition(self, range):
        if isinstance(range, str):
            return self._jmadSequenceDefinition.getRangeDefinition(range)
        else:
            return range

    @property
    def name(self):
        return self._jmadSequenceDefinition.getName()

    @property
    def ranges(self):
        return [str(r) for r in self._jmadSequenceDefinition.getRangeDefinitions()]

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name + str(self.ranges)

    def _repr_html_(self):
        html = '<em>' + self.name + '</em> with ranges:'
        html += '<ul style="margin-top:0">'
        for r in self.ranges:
            html += '<li>' + r + '</li>'
        html += '</ul>'
        return html


class MatchResult(object):
    def __init__(self, jmadMatchResult):
        self._jmadMatchResult = jmadMatchResult

    @property
    def final_penalty(self):
        return self._jmadMatchResult.getFinalPenalty()

    @property
    def vary_results(self):
        return {p.getName(): p.getFinalValue() for p in self._jmadMatchResult.getVaryParameterResults()}

    @property
    def constraint_results(self):
        return {re.sub('=\{.*?\}', '', str(p.getConstraint())): p.getFinalValue()
                for p in self._jmadMatchResult.getConstraintParameterResults()}

    def __str__(self):
        return 'Match Result ->> final penalty: ' + str(self.final_penalty)

    def __repr__(self):
        return 'Match Result ->> final penalty: ' + str(self.final_penalty) + '\n' + \
               '   Vary Results: ' + repr(self.vary_results) + '\n' + \
               '   Constraint Results: ' + repr(self.constraint_results) + '\n'


class Strengths(MutableMapping):
    def __init__(self, jmadStrengthVarSet):
        self._jmadStrengthVarSet = jmadStrengthVarSet

    def _jmadStrength(self, k):
        jmadStrength = self._jmadStrengthVarSet.getStrength(k)
        if jmadStrength is None:
            raise KeyError('Invalid Strength Name: ' + k)
        return jmadStrength

    def __getitem__(self, k):
        return self._jmadStrength(k).getValue()

    def __setitem__(self, k, v):
        return self._jmadStrength(k).setValue(float(v))

    def __delitem__(self, k):
        raise NotImplementedError('Deletion of Strengths is not supported')

    def __iter__(self):
        for s in self._jmadStrengthVarSet.getStrengths():
            yield s.getName()

    def __len__(self):
        return self._jmadStrengthVarSet.getStrengths().size()

    def _ipython_key_completions_(self):
        return [s.getName() for s in self._jmadStrengthVarSet.getStrengths()]

    def __repr__(self):
        s = 'Strengths(\n'
        for k, v in self.items():
            s += '   ' + k + ': ' + str(v) + '\n'
        s += ')'
        return s

    def __str__(self):
        return self.__repr__()

    def _repr_html_(self):
        s = '<table>'
        for k, v in self.items():
            s += '<tr><td><strong>' + k + '</strong></td><td>' + str(v) + '</td>'
        s += '</table>'
        return s


class Elements(Mapping):
    def __init__(self, elements):
        from .element import from_jmad
        if type(elements) is cern.accsoft.steering.jmad.domain.machine.Range:
            self._elementList = [from_jmad(e) for e in elements.getElements()]
        elif type(elements) is list:
            self._elementList = elements
        else:
            raise ValueError('Expecting either a list of elements or a JMad Range')
        self._nameDict = {}
        for i, e in enumerate(self._elementList):
            self._nameDict.setdefault(e.name, []).append(i)

    def __getitem__(self, k):
        if type(k) is int:
            return self._elementList[k]
        if type(k) is slice:
            if type(k.start) in (int, type(None)) and type(k.stop) in (int, type(None)) \
                    and type(k.step) in (int, type(None)):
                return self._elementList[k]
            else:
                first_name = k.start if k.start is not None else self._elementList[0]
                last_name = k.stop if k.stop is not None else self._elementList[-1]
                first_idx = self._nameDict[first_name]
                if type(first_idx) is list:
                    first_idx = first_idx[0]
                last_idx = self._nameDict[last_name]
                if type(last_idx) is list:
                    last_idx = last_idx[-1]
                if first_idx > last_idx:
                    elements = self._elementList[first_idx:] + self._elementList[:(last_idx + 1)]
                else:
                    elements = self._elementList[first_idx:(last_idx + 1)]
                return Elements(elements)
        matchingElements = [self._elementList[i] for i in self._nameDict[k]]
        if len(matchingElements) == 1:
            return matchingElements[0]
        else:
            return matchingElements

    def __iter__(self):
        return list(self._nameDict.keys())

    def items(self):
        for e in self._elementList:
            yield (e.name, e)

    def values(self):
        for e in self._elementList:
            yield e

    def __len__(self):
        return len(self._elementList)

    def _ipython_key_completions_(self):
        return list(self._nameDict.keys())

    def __repr__(self):
        s = 'Elements(\n'
        for k, v in self.items():
            s += '   ' + k + ' :: ' + str(v) + '\n'
        s += ')'
        return s

    def __str__(self):
        return self.__repr__()

    def _repr_html_(self):
        html = '<table><thead>'
        html += '<tr><th>Element</th><th>Type</th><th>S</th><th>L</th></tr>'
        html += '</thead><tbody><tr>'
        for n, e in self.items():
            html += '<tr><td><strong>' + n + '</strong></td><td>' + e.type + '</td>'
            html += '<td>' + str(e.position) + '</td><td>' + str(e.length) + '</td>'
        html += '</table>'
        return html


class _JMadVariableRepository(object):
    def __init__(self, variable_class):
        self._jmadClass = variable_class
        self._jmadVars = {v.getName().upper(): v for v in variable_class.values()}

    def _jmad_variable(self, var):
        if isinstance(var, str):
            return self._jmadVars[var.upper()]
        else:
            return var

    def __getattr__(self, name):
        if name == '_jmadVars':
            return self._jmadVars
        if name.upper() in self._jmadVars:
            return name.upper()

    def __dir__(self):
        return self._jmadVars.keys()


MadxTwissVariable = _JMadVariableRepository(cern.accsoft.steering.jmad.domain.var.enums.MadxTwissVariable)
MadxGlobalVariable = _JMadVariableRepository(cern.accsoft.steering.jmad.domain.var.enums.MadxGlobalVariable)

TfsResult = namedtuple('TfsResult', ['summary', 'data'])


def _jmad_TfsResult_to_pandas(tfs_result):
    summ = tfs_result.getSummary()
    summ_dict = {}
    for var in summ.getKeys():
        vt = str(summ.getVarType(var))
        if vt == 'STRING':
            val = summ.getStringValue(var)
        elif vt == 'DOUBLE':
            val = summ.getDoubleValue(var).doubleValue()
        else:
            val = None
        summ_dict[var] = val
    names = np.array(Iterables.toArray(tfs_result.getStringData('NAME'), java.lang.String().getClass())[:])
    result_df = pd.DataFrame(index=names)
    for var in tfs_result.getKeys():
        if var == 'NAME':
            continue
        vt = str(tfs_result.getVarType(var))
        if vt == 'STRING':
            col = np.array(Iterables.toArray(tfs_result.getStringData(var), java.lang.String().getClass())[:])
        elif vt == 'DOUBLE':
            col = np.array(Doubles.toArray(tfs_result.getDoubleData(var))[:])
        else:
            col = None
        result_df[var] = col
    return TfsResult(summary=summ_dict, data=result_df)


def _unbox_double(v):
    if v is None:
        return None
    else:
        return v.doubleValue()


def _unbox_integer(v):
    if v is None:
        return None
    else:
        return v.intValue()


def _unbox_boolean(v):
    if v is None:
        return None
    else:
        return True if v.booleanValue() == 1 else False
